### Persefone.jl - a model of agricultural landscapes and ecosystems in Europe.
###
### This file contains all macros implementing the domain-specific language
### for defining species.
###

## Note that this DSL consists of lots of deeply nested macro calls, which is a
## known tricky issue in Julia (https://github.com/JuliaLang/julia/issues/23221,
## https://github.com/p-i-/MetaGuideJulia/wiki#example-swap-macro-to-illustrate-esc).
## Hence all the `esc`apes in the following code - be careful when modifying!

"""
    @species(name, body)

A macro used to create new species definitions for the nature model.
This is effectively a simple domain-specific language, establishing a
custom syntax to describe species' biology:

```julia
@species name begin

    @initialise(@habitat(...))
    speciesvar1 = 3.14
    ...

    @phase phase1 begin
        ...
    end
end
```

The definition body (enclosed in the begin/end block) has two sections.
First comes a call to [`@initialise`](@ref), and optionally a list of
species-specific parameters, which are assigned just like normal variables.
Second come one or more phase definitions, that describe the behaviour
of the species during various parts of its life cycle. (See the documentation
to [`@initialise`](@ref) and [`@phase`](@ref) for details).

Code in a species definition block can access the rest of the model using
the `model` variable (an object of type `AgentBasedModel`).
"""
macro species(name, body)
    quote
        Core.@__doc__ function $(esc(name))($(esc(:model))::AgentBasedModel)
            # create internal variables and execute the definition body
            $(esc(:name)) = string($(QuoteNode(name)))
            $(esc(:phase)) = ""
            $(esc(body))
            # extract and process the local variables into a species dict
            vardict = Base.@locals
            speciesdict = Dict{String,Any}()
            for k in keys(vardict)
                speciesdict[string(k)] = vardict[k]
            end
            delete!(speciesdict, "model")
            delete!(speciesdict, $(string(name)))
            return speciesdict
        end
        # allow species to be defined outside of the Persefone module, but still available
        # inside it (needed by `initnature!()` and `reproduce!()`)
        (@__MODULE__() != $(esc(:Persefone))) && ($(esc(:Persefone)).$name = $(esc(name)))
    end
end

"""
    @initialise(habitatdescriptor; kwargs...)

Call this macro within the body of [`@species`](@ref). It passes the given habitat
descriptor function and keyword arguments on to [`initpopulation`](@ref) when setting
up the simulation.

Note: if this macro is not used, the variable `initialise!` must be set manually in the
species definition.
"""
macro initialise(habitatdescriptor, kwargs...)
    :($(esc(:initialise!)) = initpopulation($(esc(habitatdescriptor)); $(map(esc, kwargs)...)))
end

#TODO add an individual-level initialisation function!

"""
    @phase(name, body)

This macro is designed to be used within a species definition block (i.e. within
the body of a call to [`@species`](@ref)).

The idea behind this is that species show very different behaviour during different
phases of their lives. Therefore, `@phase` can be used define the behaviour for one
such phase, and the conditions under which the animal transitions to another phase.

`@phase` works by creating a function that will be called by the model if the
animal is in the relevant phase. When it is called, it has access to the following
variables:
- `animal` a reference to the animal itself. This provides access to `animal.age`,
    `animal.sex`, and `animal.<trait>` (where <trait> is a variable that was defined
    in the top part of the species definition body).
- `pos` gives the animal's current position as a coordinate tuple.
- `model` a reference to the model world (an object of type `AgentBasedModel`).
    This allows access to `model.date` (the current simulation date) and
    `model.landscape` (a two-dimensional array of pixels containing geographic
    information).

Several utility macros can be used within the body of `@phase` as a short-hand for
common expressions: [`@trait`](@ref), [`@setphase`](@ref), [`@respond`](@ref),
[`@kill`](@ref), [`@reproduce`](@ref), [`@neighbours`](@ref), [`@rand`](@ref),
[`@shuffle!`](@ref).

Note that the first phase that is defined in a species definition block will be
the phase that animals are assigned at birth, unless the variable `phase` is
explicitly defined by the user in the species definition block.
"""
macro phase(name, body)
    #TODO the docstrings give a lot of warnings in the log - can I fix that?
    quote
        Core.@__doc__ function $(esc(name))($(esc(:animal))::Animal, $(esc(:model))::AgentBasedModel)
            $(esc(:pos)) = $(esc(:animal)).pos
            $(esc(body))
        end
        ($(esc(:phase)) == "") && ($(esc(:phase)) = $(String(name)))
    end
end

"""
    @trait(traitname)

A utility macro to quickly access an animal's trait value.
This can only be used nested within [`@phase`](@ref).
"""
macro trait(traitname)
    #XXX This would error if called in the first part of a species definition block
    # (i.e. outside of a @phase block). Although this is specified in the documentation,
    # it is unexpected and liable to be overlooked. Can we add a third clause to
    # compensate for that?
    :($(esc(:animal)).$(traitname))
end

"""
    @setphase(newphase)

Switch this animal over to a different phase. This can only be used nested within [`@phase`](@ref).
"""
macro setphase(newphase)
    #XXX make this usable in the top part of a species definition?
    :($(esc(:animal)).phase = $(String(newphase)))
end

"""
    @respond(eventname, body)

Define how an animal responds to a landscape event that affects its current position.
This can only be used nested within [`@phase`](@ref).
"""
macro respond(eventname, body)
    quote
        if $(esc(eventname)) in $(esc(:model)).landscape[$(esc(:animal)).pos...].events
            $(esc(body))
        end
    end
end

"""
    @kill

Kill this animal (and immediately abort its current update if it dies). This is a
thin wrapper around [`kill!`](@ref), and passes on any arguments. This can only be
used nested within [`@phase`](@ref).
"""
macro kill(args...)
    :(kill!($(esc(:animal)), $(esc(:model)), $(map(esc, args)...)) && return)
end

"""
    @reproduce

Let this animal reproduce. This is a thin wrapper around [`reproduce!`](@ref), and
passes on any arguments. This can only be used nested within [`@phase`](@ref).
"""
macro reproduce(args...)
    :(reproduce!($(esc(:animal)), $(esc(:model)), $(map(esc, args)...)))
end

"""
    @migrate(arrival)

Remove this animal from the map and add it to the migrant species pool.
It will be returned to its current location at the specified `arrival` date.
This can only be used nested within [`@phase`](@ref).
"""
macro migrate(arrival)
    :(migrate!($(esc(:animal)), $(esc(:model)), $(esc(arrival))))
end

"""
    @neighbours(radius)

Return an iterator over all animals in the given radius around this animal, excluding itself.
This can only be used nested within [`@phase`](@ref).
"""
macro neighbours(radius)
    #TODO enable filtering by species
    :(nearby_animals($(esc(:animal)), $(esc(:model)), $(esc(radius))))
end

"""
    @habitat

Specify habitat suitability for spatial ecological processes.

This macro works by creating an anonymous function that takes in a model object
and a position, and returns `true` or `false` depending on the conditions
specified in the macro body.

Several utility macros can be used within the body of `@habitat` as a short-hand for
common expressions: [`@landcover`](@ref), [`@cropname`](@ref), [`@cropheight`](@ref),
[`@distanceto`](@ref), [`@distancetoedge`](@ref), [`@countanimals`](@ref).
The variables `model` and `pos` can be used for checks that don't have a macro available.

Two example uses of `@habitat` might look like this:

```julia
movementhabitat = @habitat(@landcover() in (grass agriculture soil))

nestinghabitat = @habitat((@landcover() == grass || 
                           (@landcover() == agriculture && @cropname() != "maize" &&
                            @cropheight() < 10)) &&
                          @distanceto(forest) > 20)
```

For more complex habitat suitability checks, the use of this macro can be
circumvented by directly creating an equivalent function.
"""
macro habitat(body)
    #XXX I suspect that I may have less problems with macro expansion and module
    # scoping if @habitat did not create a new function. But is there a different way?
    quote
        function($(esc(:pos)), $(esc(:model)))
            if $(esc(body))
                return true
            else
                return false
            end
        end
    end
end

##XXX Can I make sure (e.g. through `try/catch`) that the following macros
## are not called anywhere outside @habitat/@phase?

"""
    @landcover

Returns the local landcover. This is a utility wrapper that can only be used
nested within [`@phase`](@ref) or [`@habitat`](@ref).
"""
macro landcover()
    :(landcover($(esc(:pos)), $(esc(:model))))
end

"""
    @cropname

Return the name of the local croptype, or nothing if there is no crop here.
This is a utility wrapper that can only be used nested within [`@phase`](@ref)
or [`@habitat`](@ref).
"""
macro cropname()
    :(cropname($(esc(:pos)), $(esc(:model))))
end

"""
    @cropheight

Return the height of the crop at this position, or 0 if there is no crop here.
This is a utility wrapper that can only be used nested within [`@phase`](@ref)
or [`@habitat`](@ref).
"""
macro cropheight()
    :(cropheight($(esc(:pos)), $(esc(:model))))
end

"""
    @distanceto(habitat)

Calculate the distance to the closest habitat of the specified type or descriptor.
This is a utility wrapper that can only be used nested within [`@phase`](@ref)
or [`@habitat`](@ref).
"""
macro distanceto(habitat)
    :(distanceto($(esc(:pos)), $(esc(:model)), $(esc(habitat))))
end

"""
    @distancetoedge

Calculate the distance to the closest neighbouring habitat.
This is a utility wrapper that can only be used nested within [`@phase`](@ref)
or [`@habitat`](@ref).
"""
macro distancetoedge()
    :(distancetoedge($(esc(:pos)), $(esc(:model))))
end

"""
    @countanimals(species="", radius=0)

Count the number of animals of the given species in this location.
This is a utility wrapper that can only be used nested within [`@phase`](@ref)
or [`@habitat`](@ref).
"""
macro countanimals(args...)
    :(countanimals($(esc(:pos)), $(esc(:model)); $(map(esc, args)...)))
end

##TODO add movement macros
